@@ -38,6 +38,7 @@ THE SOFTWARE.  */
 #include <htslib/synced_bcf_reader.h>
 #include <htslib/vcfutils.h>
 #include <htslib/kfunc.h>
+#include <htslib/khash_str2int.h>
 #include "bcftools.h"
 #include "variantkey.h"
 #include "convert.h"
@@ -101,6 +102,9 @@ struct _convert_t
     void *dat;
     int ndat;
     char *undef_info_tag;
+    void *used_tags_hash;
+    char **used_tags_list;
+    int nused_tags;
     int allow_undef_tags;
     uint8_t **subset_samples;
 };
@@ -781,14 +785,13 @@ static void process_gp_to_prob3(convert_t *convert, bcf1_t *line, fmt_t *fmt, in
     n /= convert->nsamples;
     for (i=0; i<convert->nsamples; i++)
     {
-        float sum = 0, *ptr = (float*)convert->dat + i*n;
+        float *ptr = (float*)convert->dat + i*n;
         int j;
         for (j=0; j<n; j++)
         {
             if ( bcf_float_is_vector_end(ptr[j]) ) break;
             if ( bcf_float_is_missing(ptr[j]) ) { ptr[j]=0; continue; }
             if ( ptr[j]<0 || ptr[j]>1 ) error("[%s:%"PRId64":%f] GP value outside range [0,1]; bcftools convert expects the VCF4.3+ spec for the GP field encoding genotype posterior probabilities", bcf_seqname(convert->header,line),(int64_t) line->pos+1,ptr[j]);
-            sum+=ptr[j];
         }
         if ( j==line->n_allele )
             ksprintf(str," %f %f %f",ptr[0],0.,ptr[1]); // haploid
@@ -1205,7 +1208,35 @@ invalid:
     kputc('.', str);
 }
 
-static fmt_t *register_tag(convert_t *convert, int type, char *key, int is_gtf)
+static void _used_tags_add(convert_t *convert, int type, char *key)
+{
+    kstring_t str = {0,0,0};
+    ksprintf(&str,"%s/%s",type==T_INFO?"INFO":"FORMAT",key);
+    khash_str2int_inc(convert->used_tags_hash,str.s);
+    convert->nused_tags++;
+    convert->used_tags_list = (char**)realloc(convert->used_tags_list,sizeof(*convert->used_tags_list)*convert->nused_tags);
+    convert->used_tags_list[convert->nused_tags-1] = str.s;
+}
+
+
+#define _SET_NON_FORMAT_TAGS(function,key,...) \
+    if ( !strcmp("CHROM",key) ) { function(__VA_ARGS__, T_CHROM); } \
+    else if ( !strcmp("POS",key) ) { function(__VA_ARGS__, T_POS); } \
+    else if ( !strcmp("POS0",key) ) { function(__VA_ARGS__, T_POS0); } \
+    else if ( !strcmp("END",key) ) { function(__VA_ARGS__, T_END); } \
+    else if ( !strcmp("END0",key) ) { function(__VA_ARGS__, T_END0); } \
+    else if ( !strcmp("ID",key) ) { function(__VA_ARGS__, T_ID); } \
+    else if ( !strcmp("REF",key) ) { function(__VA_ARGS__, T_REF); } \
+    else if ( !strcmp("FIRST_ALT",key) ) { function(__VA_ARGS__, T_FIRST_ALT); } \
+    else if ( !strcmp("QUAL",key) ) { function(__VA_ARGS__, T_QUAL); } \
+    else if ( !strcmp("TYPE",key) ) { function(__VA_ARGS__, T_TYPE); } \
+    else if ( !strcmp("FILTER",key) ) { function(__VA_ARGS__, T_FILTER); } \
+    else if ( !strcmp("IS_TS",key) ) { function(__VA_ARGS__, T_IS_TS); } \
+    else if ( !strcmp("MASK",key) ) { function(__VA_ARGS__, T_MASK); } \
+    else if ( !strcmp("LINE",key) ) { function(__VA_ARGS__, T_LINE); }
+
+static void set_type(fmt_t *fmt, int type) { fmt->type = type; }
+static fmt_t *register_tag(convert_t *convert, char *key, int is_gtf, int type)
 {
     convert->nfmt++;
     if ( convert->nfmt > convert->mfmt )
@@ -1227,26 +1258,22 @@ static fmt_t *register_tag(convert_t *convert, int type, char *key, int is_gtf)
         int id = bcf_hdr_id2int(convert->header, BCF_DT_ID, key);
         if ( fmt->type==T_FORMAT && !bcf_hdr_idinfo_exists(convert->header,BCF_HL_FMT,id) )
         {
-            if ( !strcmp("CHROM",key) ) { fmt->type = T_CHROM; }
-            else if ( !strcmp("POS",key) ) { fmt->type = T_POS; }
-            else if ( !strcmp("POS0",key) ) { fmt->type = T_POS0; }
-            else if ( !strcmp("END",key) ) { fmt->type = T_END; }
-            else if ( !strcmp("END0",key) ) { fmt->type = T_END0; }
-            else if ( !strcmp("ID",key) ) { fmt->type = T_ID; }
-            else if ( !strcmp("REF",key) ) { fmt->type = T_REF; }
-            else if ( !strcmp("ALT",key) ) { fmt->type = T_ALT; }
-            else if ( !strcmp("FIRST_ALT",key) ) { fmt->type = T_FIRST_ALT; }
-            else if ( !strcmp("QUAL",key) ) { fmt->type = T_QUAL; }
-            else if ( !strcmp("FILTER",key) ) { fmt->type = T_FILTER; }
-            else if ( !strcmp("_CHROM_POS_ID",key) ) { fmt->type = T_CHROM_POS_ID; }
+            _SET_NON_FORMAT_TAGS(set_type,key,fmt)
+           else if ( !strcmp("ALT",key) ) { fmt->type = T_ALT; }
+           else if ( !strcmp("_CHROM_POS_ID",key) ) { fmt->type = T_CHROM_POS_ID; }
             else if ( !strcmp("RSX",key) ) { fmt->type = T_RSX; }
             else if ( !strcmp("VKX",key) ) { fmt->type = T_VKX; }
-            else if ( id>=0 && bcf_hdr_idinfo_exists(convert->header,BCF_HL_INFO,id) ) { fmt->type = T_INFO; }
+            else if ( id>=0 && bcf_hdr_idinfo_exists(convert->header,BCF_HL_INFO,id) )
+            {
+                fmt->type = T_INFO;
+                _used_tags_add(convert,T_INFO,key);
+            }
         }
         else if ( fmt->type==T_PBINOM )
         {
             fmt->id = bcf_hdr_id2int(convert->header, BCF_DT_ID, fmt->key);
             if ( !bcf_hdr_idinfo_exists(convert->header,BCF_HL_FMT, fmt->id)  ) error("No such FORMAT tag defined in the header: %s\n", fmt->key);
+            _used_tags_add(convert,T_FORMAT,key);
         }
         else if ( fmt->type==T_NPASS )
         {
@@ -1326,12 +1353,12 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
     kputsn(p, q-p, &str);
     if ( is_gtf )
     {
-        if ( !strcmp(str.s, "SAMPLE") ) register_tag(convert, T_SAMPLE, "SAMPLE", is_gtf);
-        else if ( !strcmp(str.s, "GT") ) register_tag(convert, T_GT, "GT", is_gtf);
-        else if ( !strcmp(str.s, "TGT") ) register_tag(convert, T_TGT, "GT", is_gtf);
+        if ( !strcmp(str.s, "SAMPLE") ) register_tag(convert, "SAMPLE", is_gtf, T_SAMPLE);
+        else if ( !strcmp(str.s, "GT") ) register_tag(convert, "GT", is_gtf, T_GT);
+        else if ( !strcmp(str.s, "TGT") ) register_tag(convert, "GT", is_gtf, T_TGT);
         else if ( !strcmp(str.s, "TBCSQ") ) 
         {
-            fmt_t *fmt = register_tag(convert, T_TBCSQ, "BCSQ", is_gtf);
+            fmt_t *fmt = register_tag(convert, "BCSQ", is_gtf, T_TBCSQ);
             fmt->subscript = parse_subscript(&q);
             if ( fmt->subscript==-1 )
             { 
@@ -1339,7 +1366,7 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             }
             else fmt->subscript++;
         }
-        else if ( !strcmp(str.s, "IUPACGT") ) register_tag(convert, T_IUPAC_GT, "GT", is_gtf);
+        else if ( !strcmp(str.s, "IUPACGT") ) register_tag(convert, "GT", is_gtf, T_IUPAC_GT);
         else if ( !strcmp(str.s, "INFO") )
         {
             if ( *q!='/' )
@@ -1355,8 +1382,9 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             while ( *q && (isalnum(*q) || *q=='_' || *q=='.') ) q++;
             if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p, &str);
-            fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
             fmt->subscript = parse_subscript(&q);
+            _used_tags_add(convert,T_INFO,str.s);
         }
         else if ( !strcmp(str.s,"PBINOM") )
         {
@@ -1366,48 +1394,34 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             while ( *q && *q!=')' ) q++;
             if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p, &str);
-            register_tag(convert, T_PBINOM, str.s, is_gtf);
+            register_tag(convert, str.s, is_gtf, T_PBINOM);
             q++;
         }
         else if ( !strcmp(str.s,"N_PASS") )
             error("N_PASS() must be placed outside the square brackets\n");
         else
         {
-            fmt_t *fmt = register_tag(convert, T_FORMAT, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_FORMAT);
             fmt->subscript = parse_subscript(&q);
         }
     }
     else
     {
-        if ( !strcmp(str.s, "CHROM") ) register_tag(convert, T_CHROM, str.s, is_gtf);
-        else if ( !strcmp(str.s, "POS") ) register_tag(convert, T_POS, str.s, is_gtf);
-        else if ( !strcmp(str.s, "POS0") ) register_tag(convert, T_POS0, str.s, is_gtf);
-        else if ( !strcmp(str.s, "END") ) register_tag(convert, T_END, str.s, is_gtf);
-        else if ( !strcmp(str.s, "END0") ) register_tag(convert, T_END0, str.s, is_gtf);
-        else if ( !strcmp(str.s, "ID") ) register_tag(convert, T_ID, str.s, is_gtf);
-        else if ( !strcmp(str.s, "REF") ) register_tag(convert, T_REF, str.s, is_gtf);
+        _SET_NON_FORMAT_TAGS(register_tag, str.s, convert, str.s, is_gtf)
         else if ( !strcmp(str.s, "ALT") ) 
         {
-            fmt_t *fmt = register_tag(convert, T_ALT, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_ALT);
             fmt->subscript = parse_subscript(&q);
         }
-        else if ( !strcmp(str.s, "FIRST_ALT") ) register_tag(convert, T_FIRST_ALT, str.s, is_gtf);
-        else if ( !strcmp(str.s, "QUAL") ) register_tag(convert, T_QUAL, str.s, is_gtf);
-        else if ( !strcmp(str.s, "FILTER") ) register_tag(convert, T_FILTER, str.s, is_gtf);
-        else if ( !strcmp(str.s, "QUAL") ) register_tag(convert, T_QUAL, str.s, is_gtf);
-        else if ( !strcmp(str.s, "IS_TS") ) register_tag(convert, T_IS_TS, str.s, is_gtf);
-        else if ( !strcmp(str.s, "TYPE") ) register_tag(convert, T_TYPE, str.s, is_gtf);
-        else if ( !strcmp(str.s, "MASK") ) register_tag(convert, T_MASK, str.s, is_gtf);
-        else if ( !strcmp(str.s, "LINE") ) register_tag(convert, T_LINE, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_CHROM_POS_ID") ) register_tag(convert, T_CHROM_POS_ID, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_PROB3") ) register_tag(convert, T_GT_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_PL_TO_PROB3") ) register_tag(convert, T_PL_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GP_TO_PROB3") ) register_tag(convert, T_GP_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_HAP") ) register_tag(convert, T_GT_TO_HAP, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_HAP2") ) register_tag(convert, T_GT_TO_HAP2, str.s, is_gtf);
-        else if ( !strcmp(str.s, "RSX") ) register_tag(convert, T_RSX, str.s, is_gtf);
-        else if ( !strcmp(str.s, "VKX") ) register_tag(convert, T_VKX, str.s, is_gtf);
-        else if ( !strcmp(str.s,"pbinom") ) error("Error: pbinom() is currently supported only with FORMAT tags. (todo)\n");
+        else if ( !strcmp(str.s, "_CHROM_POS_ID") ) register_tag(convert, str.s, is_gtf, T_CHROM_POS_ID);
+        else if ( !strcmp(str.s, "_GT_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_GT_TO_PROB3);
+        else if ( !strcmp(str.s, "_PL_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_PL_TO_PROB3);
+        else if ( !strcmp(str.s, "_GP_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_GP_TO_PROB3);
+        else if ( !strcmp(str.s, "_GT_TO_HAP") ) register_tag(convert, str.s, is_gtf, T_GT_TO_HAP);
+        else if ( !strcmp(str.s, "_GT_TO_HAP2") ) register_tag(convert, str.s, is_gtf, T_GT_TO_HAP2);
+        else if ( !strcmp(str.s, "RSX") ) register_tag(convert, str.s, is_gtf, T_RSX);
+        else if ( !strcmp(str.s, "VKX") ) register_tag(convert, str.s, is_gtf, T_VKX);
+        else if ( !strcmp(str.s,"PBINOM") ) error("Error: PBINOM() is currently supported only with FORMAT tags. (todo)\n");
         else if ( !strcmp(str.s, "INFO") )
         {
             if ( *q=='/' )
@@ -1417,14 +1431,15 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
                 while ( *q && (isalnum(*q) || *q=='_' || *q=='.') ) q++;
                 if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
                 kputsn(p, q-p, &str);
-                fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+                fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
                 fmt->subscript = parse_subscript(&q);
+                _used_tags_add(convert,T_INFO,str.s);
             }
             else
-                register_tag(convert, T_INFO, NULL, is_gtf);    // the whole INFO
+                register_tag(convert, NULL, is_gtf, T_INFO);    // the whole INFO
         }
         else if ( !strcmp(str.s, "FORMAT") )
-             register_tag(convert, T_FORMAT, NULL, 0);
+             register_tag(convert, NULL, 0, T_FORMAT);
         else if ( !strcmp(str.s,"N_PASS") )
         {
             if ( *q!='(' ) error("Could not parse the expression: %s\n", convert->format_str);
@@ -1439,12 +1454,13 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             }
             if ( q-p==0 || nopen ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p-1, &str);
-            register_tag(convert, T_NPASS, str.s, is_gtf);
+            register_tag(convert, str.s, is_gtf, T_NPASS);
         }
         else
         {
-            fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
             fmt->subscript = parse_subscript(&q);
+            _used_tags_add(convert,T_INFO,str.s);
         }
     }
     free(str.s);
@@ -1468,7 +1484,7 @@ static char *parse_sep(convert_t *convert, char *p, int is_gtf)
         q++;
     }
     if ( !str.l ) error("Could not parse format string: %s\n", convert->format_str);
-    register_tag(convert, T_SEP, str.s, is_gtf);
+    register_tag(convert, str.s, is_gtf, T_SEP);
     free(str.s);
     return q;
 }
@@ -1479,6 +1495,7 @@ convert_t *convert_init(bcf_hdr_t *hdr, int *samples, int nsamples, const char *
     convert->header = hdr;
     convert->format_str = strdup(format_str);
     convert->max_unpack = BCF_UN_STR;
+    convert->used_tags_hash = khash_str2int_init();
 
     int i, is_gtf = 0;
     char *p = convert->format_str;
@@ -1488,7 +1505,7 @@ convert_t *convert_init(bcf_hdr_t *hdr, int *samples, int nsamples, const char *
         switch (*p)
         {
             case '[': is_gtf = 1; p++; break;
-            case ']': is_gtf = 0; register_tag(convert, T_SEP, NULL, 0); p++; break;
+            case ']': is_gtf = 0; register_tag(convert, NULL, 0, T_SEP); p++; break;
             case '%': p = parse_tag(convert, p, is_gtf); break;
             default:  p = parse_sep(convert, p, is_gtf); break;
         }
@@ -1519,6 +1536,12 @@ void convert_destroy(convert_t *convert)
         if ( convert->fmt[i].destroy ) convert->fmt[i].destroy(convert->fmt[i].usr);
         free(convert->fmt[i].key);
     }
+    if ( convert->nused_tags )
+    {
+        for (i=0; i<convert->nused_tags; i++) free(convert->used_tags_list[i]);
+        free(convert->used_tags_list);
+    }
+    khash_str2int_destroy(convert->used_tags_hash);
     free(convert->fmt);
     free(convert->undef_info_tag);
     free(convert->dat);
@@ -1675,3 +1698,13 @@ int convert_max_unpack(convert_t *convert)
     return convert->max_unpack;
 }
 
+int convert_is_tag_used(convert_t *convert, char *tag)
+{
+    return khash_str2int_has_key(convert->used_tags_hash, tag);
+}
+const char **convert_list_used_tags(convert_t *convert, int *ntags)
+{
+    *ntags = convert->nused_tags;
+    return (const char **)convert->used_tags_list;
+}
+
